// copyright

/**
 * The Model Manager administrates the basic model used
 * during the execution of CSim.  Provided here are three
 * basic environments: 2d, 3d, and an architecture to allow
 * plug-ins that the user can design.
 *
 * @author Steven Bergom
 */

#ifndef _CSIM_MMANAGER_H
#define _CSIM_MMANAGER_H

class ModelManager
{
  /**
   * Public functions
   */
 public:
  ModelManager() {}
  virtual ~ModelManager() {}

  virtual double_vector computeNextPosition(double_vector &previous,
					    double_vector &inputs)=0;
  // Computes the next orientation of the model given an orientation
  // and a set of inputs.  There are no assumptions about the inputs
  // only that it is assumed that they are the "best guess."  The
  // ModelManager may massage the inputs as needed

  virtual void stopServices()=0;
  // Stops any services performed by the ModelManager

  virtual double distance(double_vector& dv1, double_vector &dv2)=0;
  // Computes distance between two configurations

  /**
   * Protected functions
   */
 protected:

  /**
   * Private functions
   */
 private:

  /**
   * Public data members
   */
 public:

  /**
   * Public data members
   */
 protected:

  /**
   * Public data members
   */
 private:
};

class MMSimple2D : public ModelManager
{
  /**
   * Public functions
   */
 public:
  MMSimple2D();
  virtual ~MMSimple2D();

  virtual double_vector computeNextPosition(double_vector &previous,
					    double_vector &inputs);
  // Computes the next orientation of the model given an orientation
  // and a set of inputs.  There are no assumptions about the inputs
  // only that it is assumed that they are the "best guess."  The
  // ModelManager may massage the inputs as needed

  virtual void stopServices();
  // Stops any services performed by the ModelManager

  virtual double distance(double_vector &dv1, double_vector &dv2);
  // Computes distance between two configurations

  /**
   * Protected functions
   */
 protected:

  /**
   * Private functions
   */
 private:

  /**
   * Public data members
   */
 public:

  /**
   * Public data members
   */
 protected:

  /**
   * Public data members
   */
 private:
};

class MMSimple3D : public ModelManager
{
  /**
   * Public functions
   */
 public:
  MMSimple3D();
  virtual ~MMSimple3D();

  virtual double_vector computeNextPosition(double_vector &previous,
					    double_vector &inputs);
  // Computes the next orientation of the model given an orientation
  // and a set of inputs.  There are no assumptions about the inputs
  // only that it is assumed that they are the "best guess."  The
  // ModelManager may massage the inputs as needed

  virtual void stopServices();
  // Stops any services performed by the ModelManager

  virtual double distance(double_vector &dv1, double_vector &dv2);
  // Computes distance between two configurations

  /**
   * Protected functions
   */
 protected:

  /**
   * Private functions
   */
 private:

  /**
   * Public data members
   */
 public:

  /**
   * Public data members
   */
 protected:

  /**
   * Public data members
   */
 private:
};

typedef enum { MPSRunning, MPSPluginNotFound, MPSPipeFailure, MPSError,
               MPSNotRunning } MPStatus;
typedef enum { CSShutdown, CSComputeNextState, CSInit,
               CSComputeDistance } CSimPluginSignal;

extern void csim_main(int argc, char **argv);
#define MAIN() int main(int argc, char **argv) { csim_main(argc, argv); return 0; }

class MMPlugIn : public ModelManager
{
  /**
   * Public functions
   */
 public:
  MMPlugIn(const char *plug_in_name);
  // Assert plug_in_name is the name of a valid executable
  // in the current execution path
  virtual ~MMPlugIn();

  virtual double_vector computeNextPosition(double_vector &previous,
					    double_vector &inputs);
  // Computes the next orientation of the model given an orientation
  // and a set of inputs.  There are no assumptions about the inputs
  // only that it is assumed that they are the "best guess."  The
  // ModelManager may massage the inputs as needed

  virtual void stopServices();
  // Stops any services performed by the ModelManager

  virtual double distance(double_vector &dv1, double_vector &dv2);
  // Computes distance between two configurations

  virtual MPStatus checkStatus();
  // Returns the current status of the plugin

  /**
   * Protected functions
   */
 protected:

  /**
   * Private functions
   */
 private:
  MMPlugIn(const MMPlugIn &mp) {}
  // Copy constructor, bad!

  /**
   * Public data members
   */
 public:

  /**
   * Public data members
   */
 protected:
  char                *plugin_name;
  int                  read_plugin, write_plugin;
  int                  plugin_read, plugin_write;
  MPStatus             current_status;
  int                  plugin_pid;

  /**
   * Public data members
   */
 private:
};

// Plugin architecture borrowed from The GIMP(tm)
// (Hey, if it works, use it!)
typedef struct {
  // The name of the plugin
  char *plugin_name;

  // The number of dimensions that it needs
  int dimensions;

  // The function to call on exit of the plugin
  void (*quit_func)(void);

  // The function to call on initialization of the plugin
  void (*init_func)(void);

  // The function to call to compute distance between vectors
  double (*distance_func)(double_vector &dv1,
			  double_vector &dv2);

  // The function that computes the next orientation
  double_vector (*run_func)(double_vector &pre,
			    double_vector &inputs);
} CSPluginInfo;

extern CSPluginInfo CSIM_PLUGIN;

class MMPlugInDL : public ModelManager
{
  /**
   * Public functions
   */
 public:
  MMPlugInDL(const char *plug_in_name);
  // Assert plug_in_name is the name of a valid executable
  // in the current execution path
  virtual ~MMPlugInDL();

  virtual double_vector computeNextPosition(double_vector &previous,
					    double_vector &inputs);
  // Computes the next orientation of the model given an orientation
  // and a set of inputs.  There are no assumptions about the inputs
  // only that it is assumed that they are the "best guess."  The
  // ModelManager may massage the inputs as needed

  virtual void stopServices();
  // Stops any services performed by the ModelManager

  virtual double distance(double_vector &dv1, double_vector &dv2);
  // Computes distance between two configurations

  /**
   * Protected functions
   */
 protected:

  /**
   * Private functions
   */
 private:
  MMPlugInDL(const MMPlugIn &mp) {}
  // Copy constructor, bad!

  /**
   * Public data members
   */
 public:

  /**
   * Public data members
   */
 protected:
  char                *plugin_name;
  MPStatus             current_status;
  void                *lib_link;
  CSPluginInfo        *funcs; 

  /**
   * Public data members
   */
 private:
};


#endif // _CSIM_MMANAGER_H
