#include "csim.h"

extern int   yyparse();
extern void  yyerror(char *s);
extern FILE *yyin;
extern int yyline;

Configuration::Configuration()
{
  signal(SIGQUIT, errorHandler);
  signal(SIGFPE,  errorHandler);
  signal(SIGABRT, errorHandler);
  signal(SIGPIPE, errorHandler);
}

Configuration::Configuration(int argc, char **argv)
{
  readConfiguration(argc, argv);
}

Configuration::~Configuration()
{
  delete environment;
  delete m_manager;
}

ConfigurationError Configuration::readConfiguration(int argc, char **argv)
{
  char *input_file_name = NULL;
  FILE *input_file = NULL;
  GraphNode begin;
  SkewType st;

  verbose = TRUE;
  out = NULL;
  program_name = argv[0];
  profiling = FALSE;
  
  for (int i=1; i<argc;) 
  {
    if (argv[i][0] == '-') 
    {
      // Must be a standard option
      if (strcmp(argv[i]+1, "2d") == 0) 
      {
	// We want a 2d environment
	if (environment != NULL) 
	{
	  delete environment;
	}
	environment = new Model2d();
        i++;
	
      } 
      else if (strcmp(argv[i]+1, "3d") == 0) 
      {
	// We want a 3d environment
	if (environment != NULL) 
	{
	  delete environment;
	}
	environment = new Model3d();
	i++;
      } 
      else if (strcmp(argv[i]+1, "q") == 0) 
      {
	// Be quiet
	verbose = FALSE;
        i++;
      } 
      else if (strcmp(argv[i]+1, "of") == 0)
      {
	// Output file
	if (argv[i+1] == NULL)
        {
	  printArguments();
        }
        out = new ofstream(argv[i+1]);
	i += 2;
      }
      else if (strcmp(argv[i]+1, "h") == 0)
      {
        // Print options
        printArguments();
      }
      else if (strcmp(argv[i]+1, "profile") == 0)
      {
	// Print out profiling
	profiling = TRUE;
	i++;
      }
      else if (strcmp(argv[i]+1, "skew") == 0)
      {
	// Set the skew type
	st = (SkewType)atol(argv[i+1]);
	i+=2;
      }
      else 
      {
	fprintf(stderr, "Invalid argument: %s\n\n", argv[i]+1);
	printArguments();
      }
      
    } 
    else 
    {
      // Must be a file name
      input_file_name = argv[i];
      i++;
    }
  }    

  if (input_file_name == NULL)
  {
    // No file name was given so we try to read from stdin
    input_file = stdin;
    
  } 
  else 
  {
    // Try to open the file given
    input_file = fopen(input_file_name, "r");
    if (!input_file) {
      fprintf(stderr, "Could not open %s.  Exiting.\n", program_name);
      exit(-1);
    }
  }
  
  // Read the configuration file
  if (readConfigFile(input_file) != TRUE) 
  {
    fprintf(stderr, "Could not get configured.  Exiting\n");
    exit(-1);
  }

  // Check output
  if (out == NULL)
  {
    out = &cout;
  }

  begin.setData(begin_state);
  path = new Graph( begin );
  accept_state.setSkewType(st);

  if (verbose)
  {
    printCopyright();
  }
}

void Configuration::printCopyright()
{
  fprintf(stdout, "CSim vehicle simulator\n");
  fprintf(stdout, "Copyright Iowa State University 1999\n");
  fprintf(stdout, "programmed by Steven Bergom\n\n");
  return;
}

void Configuration::printArguments()
{
  fprintf(stderr, "Usage: %s [OPTIONS] <filename>\n", program_name);
  fprintf(stderr, "Where [OPTIONS] is:\n");
  fprintf(stderr, "\t-h\tThis help file\n");
  fprintf(stderr, "\t-q\tDo not do verbose output\n");
  fprintf(stderr, "\t-of <file>\tOutput file\n");
  exit(-1);
  return;
}

void errorHandler(int err_no)
{
  switch(err_no) {
  case SIGQUIT:
  case SIGABRT:
    fprintf(stderr, "User interrupt... ");
    break;

  case SIGFPE:
    fprintf(stderr, "Floating point exception occurred... ");
    break;

  case SIGPIPE:
    fprintf(stderr, "Broken pipe... ");
    break;

  default:
    fprintf(stderr, "Odd exception encountered (%d)...", err_no);
    break;
  }

  fprintf(stderr, "Exiting gracefully\n");

  fprintf(stderr, "Trashing the environment... \n");
  delete main_config.environment;

  fprintf(stderr, "Making what was once dynamic, static... \n");
  delete main_config.m_manager;

  fprintf(stderr, "Have a nice day :)\n");

  exit(-1);
}

bool Configuration::readConfigFile(FILE *file_handle)
{
  int ret_val;

  // Set input file
  yyin = file_handle;

  ret_val = yyparse();

  return ret_val == 0;
}

void yyerror(char *s)
{
  fprintf(stderr, "%d: %s\n", yyline, s);
}

Acceptor::~Acceptor()
{
  int cond_size=conditions.size();
  for (int i=0; i<cond_size; i++)
  {
    delete conditions[i];
  }
}

bool Acceptor::accept(double_vector conf)
{
  int accept_count=0, cond_size=conditions.size();

  for (int i=0; i<cond_size; i++)
  {
    if (conditions[i]->accept(conf[i]) == TRUE)
    {
      accept_count++;
    }
  }

  return (accept_count == cond_size);
}

SkewType Acceptor::setSkewType(SkewType st = STNoSet)
{
  SkewType old_skew;

  switch (st)
  {
  case STNoSet:
    break;

  case STSphere:
  case STCenter:
    skew_type = st;
    skew_count = 0;
    center_goal.clear();
    for (int i=0; i<conditions.size(); i++)
    {
      center_goal.push_back(conditions[i]->center());
    }
    break;

  case STGravity:
    skew_type = st;
    break;

  case STNone:
  default:
    skew_type = STNone;
    break;
    
  }

  return old_skew;
}

double_vector Acceptor::skewConf(double_vector &dv)
{
  double_vector ret_val;

  switch (skew_type)
  {
  case STSphere:
    break;

  case STCenter:
    skew_count++;
    if (skew_count == 100)
    {
      for (int i=0; i<dv.size(); i++)
      {
	ret_val.push_back(center_goal[i] == -1 ? dv[i] : center_goal[i]);
      }
      skew_count = 0;
    }
    else
    {
      ret_val = dv;
    }
    break;

  case STGravity:
    break;

  case STNone:
    ret_val = dv;
    break;
  }

  return ret_val;
}


