// copyright

#include "csim.h"

//------------------------------------------------------------------

MMSimple2D::MMSimple2D()
{
}

MMSimple2D::~MMSimple2D()
{
}

double_vector MMSimple2D::computeNextPosition(double_vector &previous,
					      double_vector &inputs)
{
  double_vector new_state;
  int dms = previous.size();
  double grow_amt = distance(previous, inputs);

  // Assume that the previous state and the set of inputs
  // have the same length
  assert(previous.size() == inputs.size());

  grow_amt  = (grow_amt > main_config.time_step ? 
	       main_config.time_step :
	       1);
  for (int i=0; i<main_config.num_dimensions; i++)
  {
    new_state.push_back(
	previous[i] + 
	grow_amt * 
	(inputs[i] - previous[i]) );
  }

  //for (int i=0; i<dms; i++) {
  //  new_state.push_back(previous[i]+inputs[i]);
  //}

  return new_state;
}

void MMSimple2D::stopServices()
{
}

double MMSimple2D::distance(double_vector &dv1, double_vector &dv2)
{
  // This just computes the euclidean distance
  double sqr_dist=0;
  
  assert( dv1.size() == dv2.size() );

  for (int i=0; i<dv1.size(); i++)
  {
    double q_dist = dv1[i] - dv2[i];
    sqr_dist += q_dist*q_dist;
  }

  return sqrt(sqr_dist);
}


//------------------------------------------------------------------


MMSimple3D::MMSimple3D()
{
}

MMSimple3D::~MMSimple3D()
{
}

double_vector MMSimple3D::computeNextPosition(double_vector &previous,
					      double_vector &inputs)
{
  double_vector new_state;
  int dms = previous.size();

  // Assume that the previous state and the set of inputs
  // have the same length
  assert(previous.size() == inputs.size());

  for (int i=0; i<dms; i++) {
    new_state.push_back(previous[i]+inputs[i]);
  }
}

void MMSimple3D::stopServices()
{
}

double MMSimple3D::distance(double_vector &dv1, double_vector &dv2)
{
  // This just computes the squared euclidean distance
  double sqr_dist=0;

  assert( dv1.size() == dv2.size() );

  for (int i=0; i<dv1.size(); i++)
  {
    double q_dist = dv1[i] - dv2[i];
    sqr_dist += q_dist*q_dist;
  }

  return sqr_dist;
}


//------------------------------------------------------------------


MMPlugIn::MMPlugIn(const char *plug_in_name)
{
  char *plugin_args[5];
  char pipe_r[10], pipe_w[10];
  int  to_plugin[2], from_plugin[2];

  if (pipe(to_plugin)   == -1 ||
      pipe(from_plugin) == -1) {
    current_status = MPSPipeFailure;
    return;
  }

  // Store the plugin pipe information
  read_plugin  = from_plugin[0];
  write_plugin = to_plugin[1];
  plugin_read  = to_plugin[0];
  plugin_write = from_plugin[1];
  
  // Make strings out of the pipe descriptors to
  // pass to the plugin as args
  sprintf(pipe_r, "%d", plugin_read);
  sprintf(pipe_w, "%d", plugin_write);

  // Prepare the argumest for executing the plugin
  plugin_args[0] = (char*)strdup(plug_in_name);
  plugin_args[1] = pipe_r;
  plugin_args[2] = pipe_w;
  plugin_args[3] = "-csim";
  plugin_args[4] = NULL;

  plugin_pid = fork();
  
  if (plugin_pid == 0) {
    // Execute the plugin, setting the status indicator
    // as necessary
    close(read_plugin);
    close(write_plugin);
    execvp(plug_in_name, plugin_args);
    exit(1);

  } else if (plugin_pid == -1) {
    current_status == MPSError;

  } else {
    int response;
    close(plugin_read);
    close(plugin_write);
    read(read_plugin, (void *)&response, INT_SIZE);
    current_status = (response == 1 ? MPSRunning : MPSError);
  }

  return;
}

MMPlugIn::~MMPlugIn()
{
  stopServices();
  free(plugin_name);
}

double_vector MMPlugIn::computeNextPosition(double_vector &previous,
					    double_vector &inputs)
{
  int dms = previous.size();
  double_vector new_state;
  double tmp;
  CSimPluginSignal cps = CSComputeNextState;

  // Assume that the previous state and the set of inputs
  // have the same length
  assert(previous.size() == inputs.size());

  // Tell the plugin what to and how much data to expect
  write(write_plugin, (void *)&cps, INT_SIZE);
  write(write_plugin, (void *)&dms, INT_SIZE);

  // Write the previous state to the pipe
  for (int i=0; i<dms; i++) {
    write(write_plugin, (void *)&previous[i], DOUBLE_SIZE);
  }

  // Write the inputs to the pipe
  for (int i=0; i<dms; i++) {
    write(write_plugin, (void *)&inputs[i], DOUBLE_SIZE);    
  }

  // Read the answer
  for (int i=0; i<dms; i++) {
    read(read_plugin, (void *)&tmp, DOUBLE_SIZE);
    new_state.push_back(tmp);
  }

  return new_state;
}

void MMPlugIn::stopServices()
{
  CSimPluginSignal cps = CSShutdown;

  if (current_status == MPSNotRunning) {
    // Everything is already shut down, so we don't need 
    // to do anything more
    return;
  }

  // Send the close signal to the plugin
  write(write_plugin, (void *)&cps, INT_SIZE);
  
  // Wait for the plugin to exit
  waitpid((pid_t)plugin_pid, NULL, WNOHANG|WUNTRACED);

  // Close the communication pipes
  close(read_plugin);
  close(write_plugin);

  // Set the status
  current_status = MPSNotRunning;
}

double MMPlugIn::distance(double_vector &dv1, double_vector &dv2)
{
  // TODO: distance for MMPlugin
  int dms = dv1.size();
  double answer;
  double tmp;
  CSimPluginSignal cps = CSComputeDistance;

  // Assume that the previous state and the set of inputs
  // have the same length
  assert(dv1.size() == dv2.size());

  // Tell the plugin what to and how much data to expect
  write(write_plugin, (void *)&cps, INT_SIZE);
  write(write_plugin, (void *)&dms, INT_SIZE);

  // Write the states to the pipe
  for (int i=0; i<dms; i++) {
    write(write_plugin, (void *)&dv1[i], DOUBLE_SIZE);
  }

  for (int i=0; i<dms; i++) {
    write(write_plugin, (void *)&dv2[i], DOUBLE_SIZE);    
  }

  // Read the answer
  read(read_plugin, (void *)&answer, DOUBLE_SIZE);

  return answer;
}

MPStatus MMPlugIn::checkStatus()
{
  return current_status;
}



//------------------------------------------------------------------

MMPlugInDL::MMPlugInDL(const char *plug_in_name)
{
  lib_link = dlopen(plug_in_name, RTLD_NOW);
  if (lib_link == NULL)
  {
    cerr << dlerror() << endl;
    exit(1);
  }

  funcs = (CSPluginInfo*)dlsym(lib_link, "CSIM_PLUGIN");

  if (funcs == NULL)
  {
    cerr << dlerror() << endl;
    exit(1);
  }

  if (funcs->init_func != NULL)
  {
    (*funcs->init_func)();
  }
}

MMPlugInDL::~MMPlugInDL()
{
  stopServices();
}

double_vector MMPlugInDL::computeNextPosition(double_vector &previous,
					      double_vector &inputs)
{
  if (funcs->run_func != NULL)
  {
    return (*funcs->run_func)(previous, inputs);
  }
  else
  {
    return previous;
  }
}

void MMPlugInDL::stopServices()
{
  if (funcs->quit_func != NULL)
  {
    (*funcs->quit_func)();
  }
  funcs = NULL;
  dlclose(lib_link);
}


double MMPlugInDL::distance(double_vector &dv1, double_vector &dv2)
{
  if (funcs->distance_func != NULL)
  {
    return (*funcs->distance_func)(dv1, dv2);
  }
  else
  {
    return 1;
  }
}


