// copyright

/**
 */

#include "csim.h"
#include <math.h>


// Constants
const int LINEAR = 0;
const double L = 0.50526316;
const double mu = 0.85;
const double epsilon = 1e-20;

const double MASS = 3518.0/32.2;
const double BASE = 100.5/12.0;
const double A = 0.45/BASE;
const double B = 0.55/BASE;
const double Izz = 0.25*MASS*(A+B)*(A+B);
const double trackf = 55.0/12.0;
const double trackr = 54.6/12.0;

const double Nf = MASS*32.2*0.55;
const double Nr = MASS*32.2*0.45;
const double Calff = 1.0;
const double Calfr = 1.0;


void quit(void);
void init(void);
double distance(double_vector &dv1, double_vector &dv2);
double_vector run(double_vector &pre, double_vector &inputs);

CSPluginInfo CSIM_PLUGIN = 
{
  "test1_plugin",    /* name          */
  3,                 /* dimensions    */
  quit,              /* quit_func     */
  init,              /* init_func     */
  distance,          /* distance_func */
  run                /* run_func      */
};

//MAIN()

void quit(void)
{
  cerr << "Received the signal to quit...\n" << flush;
}

void init(void)
{
  cerr << "Received the signal to initialize...\n" << flush;
}

double distance(double_vector &dv1, double_vector &dv2)
{
  double dist;
  int size = dv1.size();

  // Compute the Euclidian distance metric
  for (int i=0; i<size; i++)
  {
    double dv1_inf = dv1[i] - dv2[i];
    dist += dv1_inf * dv1_inf;
  }

  return sqrt(dist);
}

double_vector run(double_vector &pre, double_vector &inputs)
{
  double cur_x, cur_y, cur_turn;
  double cur_v, u, r;
  double_vector ret_val;
  double Fyf, Fyr, cospsi, sinpsi;

  double deltaf, deltar;
  double alphaf, alphar;
  double talff, talfr;
  double xibLf, xibLr;
  double xibarf, xibarr;
  double mzf, mzr;

  cur_x = pre[0];
  cur_y = pre[1];
  cur_turn = pre[2]*2*M_PI;
  u = pre[3];
  cur_v = pre[4];
  r = pre[5];

  // Calculate various variables
  sinpsi = sin(cur_turn);
  cospsi = cos(cur_turn);

  deltaf = inputs[3]-pre[3];
  deltar = inputs[4]-pre[4];
  alphaf = -deltaf + (cur_v + A*r)/u;
  alphar = -deltar + (cur_v - B*r)/u;
  talff = tan(fabs(alphaf));
  talfr = tan(fabs(alphar));

  xibLf = mu*Nf/(2.0*Calff*talff + epsilon);
  xibLr = mu*Nr/(2.0*Calfr*talfr + epsilon);

  xibarf = L*xibLf;
  xibarr = L*xibLr;

  if (xibLf >= 1.0)
  {
    Fyf = Calff*talff;
    mzf = L*Calff*talff/6.0;
  }
  else
  {
    Fyf = mu*Nf*(1.0-0.5*xibLf);
    mzf = mu*Nf*xibarf*(3.0-2.0*xibLf)/12.0;
  }

  if (xibLr >= 1.0)
  {
    Fyr = Calfr*talfr;
    mzr = L*Calfr*talfr/6.0;
  }
  else
  {
    Fyr = mu*Nr*(1.0-0.5*xibLr);
    mzr = mu*Nr*xibarr*(3.0 - 2.0*xibLr)/12.0;
  }

  // State transition equations
  u = 0.0;
  cur_v = -u*r + (Fyf + Fyr)/MASS;
  r = (Fyf*A - Fyr*B)/Izz;
  cur_turn = r;
  cur_x = u*cospsi - cur_v*sinpsi;
  cur_y = u*sinpsi + cur_v*cospsi;

  

  // Prepare ret_val and return
  ret_val.push_back(cur_x);
  ret_val.push_back(cur_y);
  ret_val.push_back(cur_turn);
  ret_val.push_back(u);
  ret_val.push_back(cur_v);
  ret_val.push_back(r);

  return ret_val;
}

