#include "csim.h"

#define MAX_STEP   0.05
#define EPSILON    10e-10

Configuration main_config;

void initMainConfig(void);
void rrtSearch(void);
bool computeIntersection(double_vector &dv1, double_vector &dv2);

extern OutputFuncs output;

void printProfiling(void);

#define TOTAL_NODES_INSERTED 0
#define INTERSECTIONS        1
#define TOTAL_ITERATIONS     2
#define NUM_DATA             3

unsigned long data_collection[NUM_DATA];

int main(int argc, char **argv)
{
  main_config.readConfiguration(argc, argv);

  //initMainConfig();

  (*output.init_output)();

  rrtSearch();

  //while (1);

  (*output.close_output)();

  if (main_config.profiling)
  {
    printProfiling();
  }

  return 0;
}

void initMainConfig(void)
{
  double_vector bs;
  GraphNode begin;
  char *file_name = "output2d/test_file.csm";

  bs.push_back(0.0);
  bs.push_back(0.0);
  bs.push_back(0.0);

  begin.setData(bs);
  main_config.path        = new Graph( begin );

  main_config.program_name = NULL;
  main_config.begin_state = bs;
  //main_config.m_manager   = new MMSimple2D();
  //main_config.m_manager   = new MMPlugIn("test1_plugin");
  main_config.m_manager   = new MMPlugInDL("libtest2.so");
  main_config.num_its     = 10000;
  main_config.verbose     = TRUE;
  main_config.path        = new Graph( begin );
  main_config.e_type      = ET2d;
  main_config.time_step   = MAX_STEP;
  main_config.environment = new Model2d(file_name, MFTCsimNative);
  main_config.num_dimensions = 3;

  main_config.accept_state.addCondition(new ECRange(0.85, 1.0));
  main_config.accept_state.addCondition(new ECRange(0.85, 1.0));
  main_config.accept_state.addCondition(new ECAny());
}

void rrtSearch(void)
{
  for (int t=0; t<main_config.num_its; t++)
  {
    double closest_dist, t_dist;
    GraphNode closest_conf, grown_conf;
    double_vector node_data;
    double_vector random_conf;
    double_vector grow_vec;
    double grow_amt;

    data_collection[TOTAL_ITERATIONS]++;

    if (t%100 == 0)
    {
      cerr << ".";
    }
    if (t%1000 == 0)
    {
      cerr << "|";
    }

    // Generate a random configuration
    for (int i=0; i<main_config.num_dimensions; i++)
    {
      random_conf.push_back( main_config.rs.getNextRand() );
    }

    // Skew if needed
    random_conf = main_config.accept_state.skewConf(random_conf);

    // Find the closest configuration in the graph
    closest_dist = 10e17;
    for (int i=0; i<main_config.path->getNumberNodes(); i++)
    {
      t_dist = main_config.m_manager->distance( 
		   (*main_config.path)[i].getData(),
		   random_conf );
      if (t_dist < closest_dist)
      {
	closest_dist = t_dist;
	closest_conf = (*main_config.path)[i];
      }
    }

    // Grow the given configuration
    node_data = closest_conf.getData();
    grow_vec = main_config.m_manager->computeNextPosition(
        node_data, random_conf);

    // Test for intersection with the environment
    if ( !computeIntersection( node_data, grow_vec ) )
    {
      // Add the configuration to the graph
      grown_conf.setData( grow_vec );
      main_config.path->insert(closest_conf, grow_vec);

      (*output.segment_output)(closest_conf, grown_conf);
      
      data_collection[TOTAL_NODES_INSERTED]++;

      // Test to see if we are in range for end_value
      if (main_config.accept_state.accept(grow_vec))
      {
	GraphNode gn = (*main_config.path)[main_config.path->getNumberNodes()-1];
	cout << "Found the end!!!" << endl;
	(*output.path_output)(main_config.path->getNodeLine( gn ));
	return;
      }

    } // end intersection

  } // end rrt_loop

} // end rrtSearch()

void printProfiling()
{
  cout << "Profiling information " << endl;
  cout << "Total iterations:      " << data_collection[TOTAL_ITERATIONS] << endl;
  cout << "Total intersections:   " << data_collection[INTERSECTIONS] << endl;
  cout << "Total nodes:           " << data_collection[TOTAL_NODES_INSERTED] << endl;

  return;
}

// Do a linear interpolation over the path to get discrete values
// of where the vehicle may be.  Should integrate.
bool computeIntersection(double_vector &dv1, double_vector &dv2)
{
  bool ret_val = false;
  double dist, a, b, c;
  int num_discretes;

  a = dv2[0]-dv1[0];
  b = dv2[1]-dv1[1];
  c = dv2[2]-dv1[2];
  dist = sqrt(a*a+b*b+c*c);

  // Calculate the number of discrite steps based on the
  // length of the vector
  num_discretes = (int)(dist/0.01);

  // Get the increment value
  a /= num_discretes;
  b /= num_discretes;
  c /= num_discretes;

  for (int i=0; i<num_discretes; i++)
  {
    double_vector test_vec;
    test_vec.push_back(i*a+dv1[0]);
    test_vec.push_back(i*b+dv1[1]);
    test_vec.push_back(i*c+dv1[2]);    
    if ( main_config.environment->intersect(test_vec) )
    {
      return true;
    }
  }

  return ret_val;
}
