// copyright

/**
 * This is a simple plugin to test the architecture and
 * make sure that it is doing what it is supposed to.  It
 * will only allow rotations of at most PI/6.
 *
 */

#include "csim.h"

#ifndef M_PI
#define M_PI 3.14159267
#endif
 
#define THETA_MAX  0.10
#define FORCE      0.1
#define ATTITUDE   0.05
#define FUEL_MASS  1.0
#define THRUST_USE 0.01
#define ATT_USE    0.005
#define MASS       1.0
#define TIME_STEP  0.1

double grav_const;
vector<double_vector> masses;


void hover_quit(void);
void hover_init(void);
double hover_dist(double_vector &dv1, double_vector &dv2);
double_vector hover_run(double_vector &pre, double_vector &inputs);

CSPluginInfo CSIM_PLUGIN = 
{
  "hover_plugin",    /* name          */
  3,                 /* dimensions    */
  hover_quit,        /* quit_func     */
  hover_init,        /* init_func     */
  hover_dist,        /* distance_func */
  hover_run          /* run_func      */
};

//MAIN()

void hover_quit(void)
{
  //cerr << "Received the signal to quit...\n" << flush;
}

void hover_init(void)
{
}

double hover_dist(double_vector &dv1, double_vector &dv2)
{
  double dist, dv1_inf;
  int size = dv1.size()-1;

  // Compute the Euclidian distance metric
  for (int i=0; i<3; i++)
  {
    dv1_inf = dv1[i] - dv2[i];
    dist += dv1_inf * dv1_inf;
  }

  return sqrt(dist);
}

/*
 * Move the hovercraft according to the inputs
 */
double_vector transition(double_vector pre, double_vector inputs)
{
  double cur_x, cur_y, cur_t;
  double cur_x_dot, cur_y_dot, cur_t_dot;
  double thrust, t_no_trig;
  double new_x, new_y, new_theta;
  double new_x_dot, new_y_dot, new_theta_dot;
  double_vector new_state;
  double mass, use_mass;

  cur_x = pre[0];
  cur_y = pre[1];
  cur_t = pre[2];
  cur_x_dot = pre[3];  
  cur_y_dot = pre[4];
  cur_t_dot = pre[5];
  mass = pre[6];

  use_mass = THRUST_USE*inputs[0] + ATT_USE*inputs[2];
  if (use_mass > mass)
  {
    // Not enough fuel to complete maneuver
    new_theta_dot = cur_t_dot;
    new_x_dot = cur_x_dot;
    new_y_dot = cur_y_dot;
  }
  else
  {
    // Have enough fuel for maneuver
    mass -= use_mass;
    new_theta_dot = cur_t_dot + inputs[1]*ATTITUDE/MASS;
    thrust = inputs[0];

    new_theta = cur_t + new_theta_dot*TIME_STEP;
    t_no_trig = thrust*FORCE/(MASS+mass);
    new_x_dot = t_no_trig*cos(new_theta*2*M_PI) + cur_x_dot;
    new_y_dot = t_no_trig*sin(new_theta*2*M_PI) + cur_y_dot;
  }

  new_x = new_x_dot*TIME_STEP + cur_x;
  new_y = new_y_dot*TIME_STEP + cur_y;

  new_state.push_back(new_x);
  new_state.push_back(new_y);
  new_state.push_back(new_theta);
  new_state.push_back(new_x_dot);
  new_state.push_back(new_y_dot);
  new_state.push_back(new_theta_dot);
  new_state.push_back(mass);

  return new_state;
}

/*
 * Try all input combinations to determine the one that gets
 * us closer to the random configuration.
 */
double_vector findBestInput(double_vector pre, double_vector rand_conf)
{
  double_vector u_best;
  double        d_best = 1e20;

  for (double thrust = 0; thrust<=1; thrust+=1)
  {
    for (double theta=-1; theta<=1; theta+=1.0)
    {
      double dist;
      double_vector inputs;
      double_vector new_conf;

      inputs.push_back(thrust); inputs.push_back(theta);
      new_conf = transition(pre, inputs);
      dist = hover_dist(rand_conf, new_conf);

      if (dist < d_best)
      {
	d_best = dist;
	u_best = inputs;
      }
    }
  }

  return u_best;
}

/*
 * Find the best input and generate a new configuration.
 */
double_vector hover_run(double_vector &pre, double_vector &inputs)
{
  double_vector new_vec;
  double_vector best_input = findBestInput(pre, inputs);
  new_vec = transition(pre, best_input);
  return new_vec;
}
