//
// December 1998
// Steve LaValle
// Iowa State University
//
// Initial experiments with kinodynamic planning with cars 12/98

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "csim.h"

const double INFINITY = 1.0e40;
const double PI = 3.1415926535897932385;
const int STATE_DIM = 5;

inline double sqr(double x) { return x*x; }
inline double min(double x, double y) { return ((x<y) ? x : y); }

// Initial and goal states
double LS[5];  /* Lower limit on each state variable */
double US[5];  /* Upper limit on each state variable */

const double MASS        = 100.0;     // slugs
const double CAF         = 17000.0;   // Front cornering stiffness
const double CAR         = 20000.0;   // Rear cornering stuffness
const double ADIST       = 4.0;       // Mass center to front tires - feet
const double BDIST       = 5.0;       // Mass center to rear tires - feet
const double IZZ         = 1600.0;    // yaw moment of interia - ft slugs^2
const double SPEED       = 88.0;      // speed feet/sec
const double WORLD_SCALE = 0.1;       // Feet per world unit
const double DT          = 0.1;       // Time increment in seconds 

const double MIN_ACTION  = -0.6;      // Steering angle in radians
const double MAX_ACTION  = 0.6;       // Steering angle in radians
const double NUM_ACTIONS = 25;        // Quantization of possible inputs

// Scale axes to effectively take Euclidean distance in an N-cube
double state_distance_metric(double_vector s1, double_vector s2)
{
  double d = 0.0;
  int i;

  for (i = 0; i < STATE_DIM; i++) {
    d += sqr(s1[i] - s2[i]);
  }

  // Note: Not taking square root here!
  return d;
}

// This is xdot = f(x,u) for a 5dof state-space, and 1dof input
//  Model taken from Jim Bernard...
// alphaf = (v + a*r)/u - deltaf
// alphar = (v-b*r)/u
// FYF = -CAF*alphaf
// FYR = -CAR*alphar
// vdot = -u*r  + (FYF + FYR)/m
// rdot = (FYF*a - FYR*b)/Izz
// Xdot = u*cos(psi) - v*sin(psi)
// Ydot = u*sin(spi) + v*cos(psi)
// psidot = r
// 
// m = mass of car, say about 100 slugs
// CAF = front cornering stiffness in pounds per radiaof the tires, say about
// 17000
// CAR = rear cornering stiffness, say about 20000
// a is dist from mass center to front tires, say 4 feet
// b is dist from mass center to rear tires, say 5 feet
// Izz is yaw moment of intertia, say about 1600 slug ft**2
// u is forward speed which is assumed constant, input is in feet/sec
// delta is your input, it is the steer angle of the tires in radians.

double_vector state_transition_equation(double_vector s, double delta)
{
  double alphaf, alphar, fyf, fyr, v, r, psi;
  double_vector ds;

  v = s[0]; r = s[1]; psi = s[4];

  alphaf = (v + ADIST * r) / SPEED - delta;
  alphar = (v - BDIST * r) / SPEED;
  fyf = -CAF * alphaf;
  fyr = -CAR * alphar;

  /* Transfer the velocity */
  ds.push_back(-SPEED * r  + (fyf + fyr) / MASS);
  ds.push_back((fyf * ADIST - fyr * BDIST) / IZZ);
  ds.push_back(SPEED * cos(psi) - v * sin(psi));
  ds.push_back(SPEED * sin(psi) + v * cos(psi));
  ds.push_back(r);

  //cout << "s: " << s << "  Ds: " << ds << "  delta: " << delta << "\n";

  return ds;
}


// 4th order  h = time step
// The arithmetic makes use of LEDA vector operator overloads!
/*
vector runge_kutta(double_vector s, double u, double h)
{
  double_vector k1, k2, k3, k4;
  double_vector ns;
  double_vector tmp;
  double calc_tmp;
  
  k1 = state_transition_equation(s,u);

  calc_tmp = 0.5*h;
  for (int i=0; i<STATE_DIM; i++)
  {
    tmp.push_back(s[i]+calc_tmp*k1[i]);
  }
  k2 = state_transition_equation(tmp, u);

  tmp.clear();
  for (int i=0; i<STATE_DIM; i++)
  {
    tmp.push_back(s[i]+calc_tmp*k2[i]);
  }
  k3 = state_transition_equation(tmp, u);
  tmp.clear();

  for (int i=0; i<STATE_DIM; i++)
  {
    tmp.push_back(s[i]+h*k3[i]);
  }
  k4 = state_transition_equation(s + h*k3,u);

  for (int i=0; i<STATE_DIM; i++)
  {
    ns.push_back( s[i] + (h/6.0)*(k1[i] + 2.0*k2[i] + 2.0*k3[i] + k4[i]) );
  }

  return ns;
}
*/

init()
{
  // Define the state defaults
  // These are currently base on the consumer union course
  //IS[0] = 0.0; IS[1] = 0.0; IS[2] = 5.0; IS[3] = -540.0; IS[4] = 0.0;
  //GS[0] = 0.0; GS[1] = 0.0; GS[2] = 995.0; GS[3] = -540.0; GS[4] = 0.0;
  //  LS[0] = -50.0; LS[1] = -5.0; LS[2] = 0.0; LS[3] = -750.0; LS[4] = 0.0;
  //  US[0] = 50.0; US[1] = 5.0; US[2] = 1000.0; US[3] = -500.0; US[4] = 2.0*PI;
  LS[0] = -50.0; LS[1] = -5.0; LS[2] = 0.0; LS[3] = -1000.0; LS[4] = 0.0;
  US[0] = 50.0; US[1] = 5.0; US[2] = 1000.0; US[3] = 0.0; US[4] = 2.0*PI;
}

CSPluginInfo CSIM_PLUGIN = 
{
  "test2_plugin",            /* name          */
  5,                         /* dimensions    */
  quit,                      /* quit_func     */
  init,                      /* init_func     */
  state_distance_metric,     /* distance_func */
  state_transition_equation  /* run_func      */
};
